
#define WIN32_LEAN_AND_MEAN

// TODO: Remove these at some point!
#pragma warning(push, 0)

// This is placed here to resolve a Point type ambiguity on MacOS
#include <ossia/network/base/node_attributes.hpp>
#include <ossia/network/oscquery/oscquery_server.hpp>

#include "OscQueryAddress.h"

#include <ossia/detail/config.hpp>
#include <ossia/network/domain/domain.hpp>
#include <ossia/network/common/debug.hpp>
#include <ossia/network/oscquery/oscquery_mirror.hpp>
#include <ossia/network/base/parameter_data.hpp>
#include <ossia/network/generic/generic_device.hpp>

#pragma warning(pop)

#if JUCE_WINDOWS
#include <winsock2.h>   // this must come first to prevent errors with MSVC7
#include <WS2tcpip.h>
#else
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#endif

#include "../../Program/TwoFactory.h"
#include "../../Program/ModelStateBroadcaster.h"
#include "../../Program/ScopedLockingValueSource.h"
#include "../ZombieContainer.h"
#include "../../Program/CommandManager.h"
#include "../../Operations/ActionMessageWrapper.h"
#include "../../Operations/Commands/UndoableOperation.h"

#include "../../Program/Controller.h"
#include "../../OSC/OscMessageForwarder.h"

#include "../Locations/OscQueryLocation.h"
#include "../Node.h"
#include "../NodeRoot.h"
#include "../TypeTagString.h"
#include "../TypeTag.h"

#include "../../Program/Selector.h"
#include "../../Program/NodeTreeBuilder.h"

#include "../../Operations/Commands/NodeOperation.h"

using namespace ossia;
using namespace ossia::net;

OSCQueryAddress::OSCQueryAddress(Value& inOutgoingOscQueryPort,
                                 Value& inIncomingPort,
                                 String id,
                                 Controller* c,
                                 bool createContainers,
                                 Creator creator) : OscAddress(id, c, createContainers, creator),
                                                    Thread("OSC_QUERY_THREAD_" + id),
                                                    m_OutgoingPortOscQuery(new ScopedLockingValueSource()),
                                                    m_IncomingPort(new ScopedLockingValueSource()) {
    m_IsBeingDeleted = false;

    setEchoOnLoad(true);

    startThread();

    if (createContainers) {
        m_ChosenZeroconfDeviceName = "Manual";

        triggerQuery();
    }

    m_OutgoingPortOscQuery = inOutgoingOscQueryPort.getValue();
    m_IncomingPort = inIncomingPort.getValue();

    m_OutgoingPortOscQuery.addListener(this);
    m_IncomingPort.addListener(this);
    m_ChosenZeroconfDeviceName.addListener(this);

    m_ConnectionStatus = false;
    m_ConnectionStatus.addListener(this);
}

OSCQueryAddress::OSCQueryAddress(const OSCQueryAddress& other) : OscAddress(other),
                                                                 m_OutgoingPortOscQuery(new ScopedLockingValueSource()),
                                                                 m_IncomingPort(new ScopedLockingValueSource()),
                                                                 Thread("OSC_QUERY_THREAD_" + this->getID().toString() + "_1") {
    m_IsBeingDeleted = other.m_IsBeingDeleted;
    m_EchoOnLoad = other.m_EchoOnLoad.getValue();

    m_OutgoingPortOscQuery = other.m_OutgoingPortOscQuery.getValue();
    m_IncomingPort = other.m_IncomingPort.getValue();

    m_OutgoingPortOscQuery.addListener(this);
    m_IncomingPort.addListener(this);
    m_ChosenZeroconfDeviceName.addListener(this);
}

OSCQueryAddress::~OSCQueryAddress() {
    disconnect();
    triggerWillBeDeletedMessage();
    m_IsBeingDeleted = true;
    close();
}

OSCQueryAddress* OSCQueryAddress::clone() const {
    return new OSCQueryAddress(*this);
}

OSCQueryAddress& OSCQueryAddress::operator=(const OSCQueryAddress& other) {
    if (this != &other) {
        OscAddress::operator=(other);
    }

    return *this;
}

void OSCQueryAddress::fetchZeroconfDevices() {
    postStatusMessage("Fetching OSC Query Zeroconf devices...");

    {
        ScopedLock zsl(zeroconfCriticalSection);
        m_ZeroconfDevices.clear();
    }

    if (ossia::net::zeroconf_supported()) {
        // TODO: This leaks! Reported.
        auto devices = ossia::net::list_oscquery_devices();

        {
            ScopedLock zsl(zeroconfCriticalSection);

            for (auto device : devices) {
                OscZeroconfDevice newDevice;
                newDevice.name = device.name;
                newDevice.port = device.port;
                newDevice.host = device.host;

                // newDevice.host.pop_back(); // Apparently trailing '.' not a bug. Keep in mind.

                getIP(newDevice);

                m_ZeroconfDevices.push_back(newDevice);
            }
        }

        ScopedLock sl(listenersCriticalSection);
        listeners.call(&Listener::zeroconfList_Changed, *this);

        postStatusMessage("OSC Query Zeroconf devices fetched.");
    }
    else {
        DBG("Zeroconf Not Supported in this libossia build");
    }
}

void OSCQueryAddress::postStatusMessage(const char* message) {
    auto mess = new EVENT_Message_Strings();
    mess->Type = EventMessageType::Notification;
    mess->Message = message;
    mess->MessageTime = Time::getCurrentTime().toString(false, true, true, true);
    r_Controller->postEventMessage(mess);
}

void OSCQueryAddress::getIP(OscZeroconfDevice& newDevice) {
    try {
        
        struct addrinfo* info = nullptr;

        struct addrinfo hints;
        zerostruct(hints);
        hints.ai_family = AF_UNSPEC;
        hints.ai_socktype = SOCK_DGRAM;
        hints.ai_flags = AI_NUMERICSERV;
        auto status = getaddrinfo(newDevice.host.c_str(), String(newDevice.port).toRawUTF8(), &hints, &info);

        if (status == 0) {
            for (struct addrinfo* p = info; p != nullptr; p = p->ai_next) {
                struct in_addr* addr = nullptr;

                char host[NI_MAXHOST], port[NI_MAXSERV];
                int err;
                if ((err = getnameinfo(p->ai_addr, p->ai_addrlen,
                    host, sizeof host, port, sizeof port,
                    NI_NUMERICHOST)) == 0) {
                    auto message = "host: " + String(host) + ", " + String(p->ai_family) + ", " + String(port) + '\n';
                    postStatusMessage(message.toRawUTF8());

                    if (p->ai_family == AF_INET) {
                        auto ipv = (struct sockaddr_in*)p->ai_addr;
                        addr = &(ipv->sin_addr);

                        char ipstr[INET_ADDRSTRLEN];

                        inet_ntop(p->ai_family, addr, ipstr, sizeof ipstr);

                        String ip(ipstr);
                        newDevice.ip = ip.toStdString();

                        if (ip.isEmpty()) {
                            jassertfalse;
                        }
                    }
                    else {
                        auto ipv6 = (struct sockaddr_in6*)p->ai_addr;
                        addr = (struct in_addr*)&(ipv6->sin6_addr);

                        // TODO: Use also IPv6
                    }
                }
                else {
                    DBG("getnameinfo Error: " + String(gai_strerror(err)) + '\n');

                    newDevice.ip = "0.0.0.0";
                }
            }
        }
        else {
            newDevice.ip = "0.0.0.0";
            DBG("getaddrinfo Error: " + String(gai_strerror(status)) + '\n');
        }

        freeaddrinfo(info);
        
    }
    catch ([[maybe_unused]] const std::exception& e) {
        DBG("Exception while fetching IP's from hostnames: " + String(e.what()));
        jassertfalse;
    }
}

void OSCQueryAddress::on_message_cb(const ossia::net::parameter_base& p) {
    if (m_HasListen)
        return;

    auto& n = p.get_node();

    auto val = p.value();

    String tts;

    std::unique_ptr<Array<var>> valuesToSend;
    valuesToSend.reset(bridgeMessageValues(val, tts));

    StringArray valueString;
    for (int i = 0; i < valuesToSend->size(); i++) {
        valueString.add(valuesToSend->operator[](i));
    }

    std::string addressPattern = n.osc_address();

    auto messageStrigs = new OSCMessageStrings();
    messageStrigs->addressPattern = addressPattern;
    messageStrigs->typeTag = tts;
    messageStrigs->valueString = valueString.joinIntoString(", ");
    messageStrigs->sourceDescription = r_OSCQueryLocation->getID().toString();
    r_Controller->postConsoleMessage(messageStrigs, ConsoleMessageType::Outgoing);

    const osc::uint64 timetag = 1;

    /*
    String valString;
    valString << "[message] " << addressPattern
        << " <- " << value_to_pretty_string(val);
    DBG("Value Update on_message_cb: " + valString);
    */

    r_Controller->getOscMessageForwarder()->processMessage(addressPattern,
                                                           tts,
                                                           valuesToSend.get(),
                                                           r_OSCQueryLocation,
                                                           timetag,
                                                           MessageDestinations::ai_and_location,
                                                           MessageOrigin::outside);
}

void OSCQueryAddress::on_unhandled_message_cb(const std::string addressPattern, const ossia::value& val) {
    String tts;
    std::unique_ptr<Array<var>> valuesToSend;
    valuesToSend.reset(bridgeMessageValues(val, tts));

    StringArray valueString;
    for (int i = 0; i < valuesToSend->size(); i++) {
        valueString.add(valuesToSend->operator[](i));
    }

    auto messageStrigs = new OSCMessageStrings();
    messageStrigs->addressPattern = addressPattern;
    messageStrigs->typeTag = tts;
    messageStrigs->valueString = valueString.joinIntoString(", ");
    messageStrigs->sourceDescription = r_OSCQueryLocation->getID().toString();
    r_Controller->postConsoleMessage(messageStrigs, ConsoleMessageType::Outgoing);

    const osc::uint64 timetag = 1;

    r_Controller->getOscMessageForwarder()->processMessage(addressPattern,
        tts,
        valuesToSend.get(),
        r_OSCQueryLocation,
        timetag,
        MessageDestinations::ai_and_location,
        MessageOrigin::outside);
}

var ossia_value_to_val(const ossia::value& val) {
    if (val.valid()) {
        const auto type = val.get_type();

        switch (type) {
        case ossia::val_type::INT: {
            var value = val.get<int>();
            return value;
            break;
        }
        case ossia::val_type::FLOAT: {
            var value = val.get<float>();
            return value;
            break;
        }
        case ossia::val_type::VEC2F: {
            jassertfalse;
            break;
        }
        case ossia::val_type::VEC3F: {
            jassertfalse;
            break;
        }
        case ossia::val_type::VEC4F: {
            jassertfalse;
            break;
        }
        case ossia::val_type::IMPULSE: {
            //auto value = val.get<float>();
            // TODO: DEAL WITH IMPULSE
            jassertfalse;
            break;
        }
        case ossia::val_type::BOOL: {
            var value = val.get<bool>();
            return value;
            break;
        }
        case ossia::val_type::STRING: {
            String value = val.get<std::string>();
            return value;
            break;
        }
        case ossia::val_type::LIST: {
            jassertfalse;
            break;
        }
        case ossia::val_type::NONE: {
            // TODO: I should deal with None too I guess, somehow?
            jassertfalse;
            break;
        }
        }
    }

    return var();
}

Array<var>* OSCQueryAddress::bridgeMessageValues(const ossia::value& val, juce::String& tts) {
    Array<var>* valuesToSend = new Array<var>();

    if (val.valid()) {
        const auto type = val.get_type();

        switch (type) {
        case ossia::val_type::INT: {
            var value = val.get<int>();
            valuesToSend->add(value);
            tts = "i";
            break;
        }
        case ossia::val_type::FLOAT: {
            var value = val.get<float>();
            valuesToSend->add(value);
            tts = "f";
            break;
        }
        case ossia::val_type::VEC2F: {
            auto vec2f = val.get<ossia::vec2f>();
            for (int i = 0; i < 2; i++) {
                auto value = vec2f[i];
                valuesToSend->add(value);
            }

            tts = "ff";
            break;
        }
        case ossia::val_type::VEC3F: {
            auto vec3f = val.get<ossia::vec3f>();
            for (int i = 0; i < 3; i++) {
                auto value = vec3f[i];
                valuesToSend->add(value);
            }

            tts = "fff";
            break;
        }
        case ossia::val_type::VEC4F: {
            auto vec4f = val.get<ossia::vec4f>();
            for (int i = 0; i < 4; i++) {
                auto value = vec4f[i];
                valuesToSend->add(value);
            }

            tts = "ffff";
            break;
        }
        case ossia::val_type::IMPULSE: {
            //auto value = val.get<float>();
            // TODO: DEAL WITH IMPULSE
            jassertfalse;
            tts = "I";
            break;
        }
        case ossia::val_type::BOOL: {
            var value = val.get<bool>();
            valuesToSend->add(value);
            tts = "B";
            break;
        }
        case ossia::val_type::STRING: {
            String value = val.get<std::string>();
            valuesToSend->add(value);
            tts = "s";
            break;
        }
        case ossia::val_type::LIST: {
            auto values = val.get<std::vector<ossia::value>>();
            for (auto ossia_val : values) {
                var value = ossia_value_to_val(ossia_val);
                valuesToSend->add(value);
            }
            break;
        }
        case ossia::val_type::NONE: {
            // TODO: I should deal with None too I guess, somehow?
            jassertfalse;
            break;
        }
        }
    }

    return valuesToSend;
}

void OSCQueryAddress::connectAndQuery() {
    if (r_OSCQueryLocation.get() == nullptr) {
        DBG("OSC QUERY LOCATION IS NULL");
        return;
    }

    auto queryHostString = getQueryHostString();
    auto inPort = getIncomingPort();
    auto inPortInt = static_cast<int>(inPort.getValue());

    bool connected = false;

    try {
        auto hostString = queryHostString.toStdString();
        auto protocol = new ossia::oscquery::oscquery_mirror_protocol(hostString, inPortInt);
        m_Device = std::make_unique<ossia::net::generic_device>(std::unique_ptr<ossia::net::protocol_base>(protocol), "Mirror");

        auto fut = m_Device->get_protocol().update_async(m_Device->get_root_node());
        auto status = fut.wait_for(std::chrono::seconds(5));

        if (status == std::future_status::ready) {
            connected = true;
        }
        else if (status == std::future_status::timeout) {
            DBG("TIMEOUT");
            connected = false;
        }
        else {
            connected = false;
        }

        parseHostInfo(protocol);
    }
    catch (const std::exception& e)
    {
        std::string what(e.what());

        DBG("CAUGHT AN EXCEPTION!: " + what);
        jassertfalse;
        return;
    }

    if (connected) {
        m_ConnectedHostString = queryHostString;

        auto& queryRootNode = m_Device->get_root_node();

        m_Device->on_node_created.connect<&OSCQueryAddress::node_created_cb>(this);
        m_Device->on_node_removing.connect<&OSCQueryAddress::node_removing_cb>(this);
        m_Device->on_parameter_created.connect<&OSCQueryAddress::param_created_cb>(this);
        m_Device->on_parameter_removing.connect<&OSCQueryAddress::param_removing_cb>(this);

        m_Device->on_message.connect<&OSCQueryAddress::on_message_cb>(this);

        m_Device->on_unhandled_message.connect<&OSCQueryAddress::on_unhandled_message_cb>(this);

        auto rootGeneric = getRootNode();

        /*
        * 1. Check if the Address is used.
        * 2. If this address is NOT used:
        * 2.1 Make new Namespace, and set as root.
        * 2.1.1 Name it after host name.
        * 2.2 If USED:
        * 2.2.1 Fill existing Namespace.
        *
        * All the above, Undoredoable.
        *
        */

        auto rootNode = dynamic_cast<NodeRoot*>(rootGeneric);

        /*
         * This clears the nodes in the namespace if they're unused.
         * I don't see why that was a good idea?
        {
            Array<AddressInstance*> addressInstances;
            r_Controller->getMC()->checkIfAddressIsUsed(this, addressInstances);

            if (addressInstances.isEmpty() && rootGeneric->getSubNodeCount() > 0) {
                int c = rootNode->getSubNodeCount();
                while (c > 0) {
                    rootNode->removeSubNode(dynamic_cast<Node*>(rootNode->getSubNode(c - 1)), true);
                    c = rootNode->getSubNodeCount();
                }
            }
        }
        */

        // TODO IF LOADING THIS IS NULL. TEMP CHECK. FIX.
        if (rootNode) {
            r_Controller->getSelector()->mute();
            r_Controller->getNodeTreeBuilder()->setUpdateRanges(true);

            buildNodeLevelTB(queryRootNode);

            r_Controller->getNodeTreeBuilder()->setUpdateRanges(false);
            r_Controller->getSelector()->unMute();
        }
        else {
            jassertfalse;
        }

        auto message = "Connected to: " + queryHostString;
        postStatusMessage(message.toRawUTF8());

        m_ConnectionStatus = true;
        m_Connected = true;
    }
    else {
        // TODO: Maybe not needed - it happens elsewhere.
        disconnect();
        auto message = "Connecting to: " + queryHostString + " failed";
        postStatusMessage(message.toRawUTF8());
        m_ConnectionStatus = false;
        m_Connected = false;
    }
}

void OSCQueryAddress::disconnect() {
    if (m_Device) {
        m_Device->get_protocol().stop(); // If I don't do this I get callbacks for messages that crash TWO...
        
        // TODO: Should this really be commented.
        m_Device = nullptr;

        auto message = "Disconnected from: " + m_ConnectedHostString;
        postStatusMessage(message.toRawUTF8());
    }
}

void OSCQueryAddress::parseHostInfo(ossia::oscquery::oscquery_mirror_protocol* protocol) {
    auto host_info = protocol->get_host_info();
    
    DBG("Host : \"" + host_info.name + "\"");

    m_HasListen = false;
    for (auto [f, s] : host_info.extensions) {
        DBG("Has: " << f << " " << (s ? "true" : "false"));
        if (f.compare("LISTEN") == 0 && s == true) {
            m_HasListen = true;
        }
    }

    if (host_info.osc_ip.has_value()) {
        std::string ip = host_info.osc_ip.value();
        DBG("Target IP: " + ip);
    }
    // VDMX Gives both IP and Port here. And, it is kinda wrong, it doesn't contain 1234 which is the real target.
    // MadMapper doesn't give port at all.

    if (host_info.osc_port.has_value()) {
        int port = host_info.osc_port.value();
        DBG("Target Port: " + port);
    }

    if (host_info.osc_transport.has_value()) {
        auto transport = host_info.osc_transport.value();
        if (transport == ossia::oscquery::host_info::osc_transport::TCP) {
            DBG("Transport: TCP");
        }
        else { // UDP
            DBG("Transport: UDP");
        }
    }
    else { // UDP
        DBG("Transport: UDP");
    }

    if (host_info.ws_ip.has_value()) {
        std::string ip = host_info.ws_ip.value();
        DBG("WebSocket IP: " + ip);
    }

    if (host_info.ws_port.has_value()) {
        int port = host_info.ws_port.value();
        DBG("WebSocket Port: " + port);
    }
}

void OSCQueryAddress::node_created_cb(ossia::net::node_base& n) {
    auto addr = n.osc_address();

    DBG("create node : " + addr);

    // This gets called with madmapper!

    r_Controller->getSelector()->mute();
    r_Controller->getNodeTreeBuilder()->setUpdateRanges(true);

    buildNode(n);

    r_Controller->getNodeTreeBuilder()->setUpdateRanges(false);
    r_Controller->getSelector()->unMute();
}

void OSCQueryAddress::param_created_cb(const ossia::net::parameter_base& p) {
    auto& n = p.get_node();
    
    auto addr = n.osc_address();

    DBG("create param: " << addr);

    // This gets called with madmapper!

    r_Controller->getSelector()->mute();
    r_Controller->getNodeTreeBuilder()->setUpdateRanges(true);

    buildNode(n);

    r_Controller->getNodeTreeBuilder()->setUpdateRanges(false);
    r_Controller->getSelector()->unMute();
}

void OSCQueryAddress::node_removing_cb(ossia::net::node_base& n) {
    
    auto addr = n.osc_address();

    DBG("removing node : " << addr);

    // TODO: Fill in with SOMETHING...?
    // Alternatives:
    // * Turn this and children to zombies.
    // * Delete
}

void OSCQueryAddress::param_removing_cb(const ossia::net::parameter_base& p) {
    auto& n = p.get_node();
    
    auto addr = n.osc_address();

    DBG("removing param : " << addr);

    // TODO: Fill in with SOMETHING.
    // Delete NODES? Maybe not a bad idea!
}

void OSCQueryAddress::buildNodeLevelTB(const ossia::net::node_base& node) {
    String incomingAddressPattern = node.osc_address();

    for (const auto& child : node.children_copy()) {
        // I should NOT only treat those that have parameters if we implement CLASS!
        //if (auto parameter = child->get_parameter())
        buildNode(*child);
    }
}

void OSCQueryAddress::buildNode(const ossia::net::node_base& n) {
    String incomingAddressPattern = n.osc_address();

    m_NewNodeMetadata.clearQuick();

    getMetadata(n, m_NewNodeMetadata);

    auto location = this->getLocation();

    jassert(location);

    if (m_NewNodeMetadata.tts.isNotEmpty()) {
        r_Controller->getNodeTreeBuilder()->processMessage(incomingAddressPattern,
                                                           m_NewNodeMetadata,
                                                           location);
    }
    else {
        DBG("TreeBuilder Skipping AP: " + incomingAddressPattern);
    }

    String print(n.get_name());

    buildNodeLevelTB(n);
}

void OSCQueryAddress::getMetadata(const ossia::net::node_base& node,
                                  NewNodeMetadata& metadata) {
    metadata.clip = false;
    metadata.behaviour = Behaviour::Discreet;
    metadata.mute = false;
    metadata.controlType = ControlType::HorisontalSlider;
    metadata.nodeDirection = NodeDirection::Bidirectional;

    // TODO: Reinstate?
    /*
    auto description = ossia::net::get_description(node);
    if (description.has_value()) {
        metadata.description = *description;
    }*/

    auto accessMode = ossia::net::get_access_mode(node); // Get/Set/Bi
    if (accessMode.has_value()) {
        if (*accessMode == ossia::access_mode::BI) {
            metadata.nodeDirection = NodeDirection::Bidirectional;
        }
        else if (*accessMode == ossia::access_mode::GET) {
            metadata.nodeDirection = NodeDirection::Source;
        }
        else if (*accessMode == ossia::access_mode::SET) {
            metadata.nodeDirection = NodeDirection::Destination;
        }
    }

    // If I get this I can make Address hierarchy.
    // Keep assertion to see if I find it.
    /*
    auto instanceBounds = ossia::net::get_instance_bounds(node);
    if (instanceBounds.has_value()) {
        jassert(instanceBounds->min_instances == instanceBounds->max_instances);
    }
    */

//  auto acceptedValues; // Interesting for drop-downs! Does modern ossia have it?

    auto parameter = node.get_parameter();
    if (parameter != nullptr) {
        processParameter(parameter, metadata);
    }
}

void OSCQueryAddress::processParameter(ossia::net::parameter_base* parameter, NewNodeMetadata& metadata) {
    // attach to callback to display value update
    parameter->add_callback([=](const ossia::value& v) {
        // TODO: Reinstate?
        /*
        String valString;
        valString << "[message] " << ossia::net::osc_parameter_string(*parameter)
        << " <- " << value_to_pretty_string(v);
        DBG("Value Update: " + valString);
        */

        auto& n = parameter->get_node();
        auto val = parameter->value();

        String tts;

        std::unique_ptr<Array<var>> valuesToSend;

        valuesToSend.reset(bridgeMessageValues(v, tts));

        StringArray valueString;
        for (int i = 0; i < valuesToSend->size(); i++) {
            valueString.add(valuesToSend->operator[](i));
        }

        auto addressPattern = n.osc_address();

        auto messageStrigs = new OSCMessageStrings();
        messageStrigs->addressPattern = addressPattern;
        messageStrigs->typeTag = tts;
        messageStrigs->valueString = valueString.joinIntoString(", ");
        messageStrigs->sourceDescription = r_OSCQueryLocation->getID().toString();
        r_Controller->postConsoleMessage(messageStrigs, ConsoleMessageType::Outgoing);

        const osc::uint64 timetag = 1;

        r_Controller->getOscMessageForwarder()->processMessage(addressPattern,
                                                               tts,
                                                               valuesToSend.get(),
                                                               r_OSCQueryLocation,
                                                               timetag,
                                                               MessageDestinations::no_trigger,
                                                               MessageOrigin::outside);
        }
    );

    auto boundingMode = parameter->get_bounding(); // Clip / NoClip
    if (boundingMode == ossia::bounding_mode::CLIP ||
        boundingMode == ossia::bounding_mode::WRAP ||
        boundingMode == ossia::bounding_mode::FOLD) {
        metadata.clip = true;
    }
    else {
        metadata.clip = false;
    }

    auto repetitionFilter = parameter->get_repetition_filter();
    if (repetitionFilter == ossia::repetition_filter::ON) {
        metadata.behaviour = Behaviour::Continuous;
    }
    else {
        metadata.behaviour = Behaviour::Discreet;
    }

    metadata.mute = parameter->get_muted();
    //bool disabled = parameter->get_disabled(); // TODO Q: What's the difference to muted?

    auto unit = parameter->get_unit();

    auto rgba = ossia::unit_t{ ossia::rgba{} };
    auto rgb = ossia::unit_t{ ossia::rgb{} };
    auto rgba_u = ossia::unit_t{ ossia::rgba_u{} };
    auto rgb_u = ossia::unit_t{ ossia::rgb_u{} };
    auto rgba8_u = ossia::unit_t{ ossia::rgba8_u{} };

    auto hsv = ossia::unit_t{ ossia::hsv{} };

    if (parameter->get_unit() == rgba) {
        metadata.controlType = ControlType::Colour_RGB_A;
    }
    else if (parameter->get_unit() == rgb) {
        metadata.controlType = ControlType::Colour_RGB_A;
    }
    else if (parameter->get_unit() == rgba_u) {
        metadata.controlType = ControlType::Colour_RGB_A;
    }
    else if (parameter->get_unit() == rgba8_u) {
        metadata.controlType = ControlType::Colour_RGB_A;
    }
    else if (parameter->get_unit() == rgb_u) {
        metadata.controlType = ControlType::Colour_RGB_A;
    }
    else if (parameter->get_unit() == hsv) {
        metadata.controlType = ControlType::Colour_HSV_A;
    }

    extractParameter(parameter, metadata);
}

void OSCQueryAddress::extractParameter(ossia::net::parameter_base* parameter, NewNodeMetadata& metadata) {
    auto valueType = parameter->get_value_type();

    auto& dom = parameter->get_domain();
    auto min = dom.get_min();
    auto max = dom.get_max();

    switch (valueType)
    {
    case ossia::val_type::IMPULSE: {
        metadata.values.add(true);
        metadata.tts = "I";

        break;
    }
    case ossia::val_type::BOOL: {
        var val(parameter->value().get<bool>());
        metadata.values.add(val);
        metadata.tts = "B";

        break;
    }
    case ossia::val_type::INT: {
        int val = parameter->value().get<int>();
        metadata.values.add(val);

        metadata.hasRanges = false;

        if (min.valid() && max.valid()) {
            metadata.minValues.add(ossia::convert<int>(min));
            metadata.maxValues.add(ossia::convert<int>(max));

            removeRangesIfMinMaxSame(min, max, metadata);
        }
        metadata.tts = "i";

        break;
    }
    case ossia::val_type::FLOAT: {
        float val = parameter->value().get<float>();
        metadata.values.add(val);
        metadata.hasRanges = false;

        if (min.valid() && max.valid()) {
            metadata.minValues.add(ossia::convert<float>(min));
            metadata.maxValues.add(ossia::convert<float>(max));

            removeRangesIfMinMaxSame(min, max, metadata);
        }

        metadata.tts = "f";

        break;
    }
    case ossia::val_type::STRING: {
        String val = parameter->value().get<std::string>();

        metadata.values.add(val);
        metadata.tts = "s";

        break;
    }
    case ossia::val_type::VEC2F: {
        auto vec2f = parameter->value().get<ossia::vec2f>();

        for (int i = 0; i < 2; i++) {
            auto val = vec2f[i];
            metadata.values.add(val);
        }

        metadata.hasRanges = false;

        if (min.valid() && max.valid()) {
            metadata.minValues.add(ossia::convert<float>(min));
            metadata.maxValues.add(ossia::convert<float>(max));

            removeRangesIfMinMaxSame(min, max, metadata);
        }

        metadata.tts = "ff";

        break;
    }
    case ossia::val_type::VEC3F: {
        auto vec3f = parameter->value().get<ossia::vec3f>();

        for (int i = 0; i < 3; i++) {
            auto val = vec3f[i];
            metadata.values.add(val);
        }

        metadata.hasRanges = false;

        if (min.valid() && max.valid()) {
            metadata.minValues.add(ossia::convert<float>(min));
            metadata.maxValues.add(ossia::convert<float>(max));

            removeRangesIfMinMaxSame(min, max, metadata);
        }

        metadata.tts = "fff";

        break;
    }
    case ossia::val_type::VEC4F: {
        auto vec4f = parameter->value().get<ossia::vec4f>();

        for (int i = 0; i < 4; i++) {
            auto val = vec4f[i];
            metadata.values.add(val);
        }

        metadata.hasRanges = false;

        if (min.valid() && max.valid()) {
            metadata.minValues.add(ossia::convert<float>(min));
            metadata.maxValues.add(ossia::convert<float>(max));

            removeRangesIfMinMaxSame(min, max, metadata);
        }

        metadata.tts = "ffff";

        break;
    }
    default:
        break;
    }
    // TODO Tuple, etc...
}

void OSCQueryAddress::removeRangesIfMinMaxSame(ossia::value& min, ossia::value& max, NewNodeMetadata& metadata) {
    if (min == max) {
        metadata.hasRanges = false;
    }
    else {
        metadata.hasRanges = true;
    }
}

void OSCQueryAddress::run() {
/*
    if (m_ZeroconfDevices.empty())
        fetchZeroconfDevices();
*/

    if (r_Controller->getModel() == nullptr) {
        DBG("MODEL IS NULL");
        return;
    }

    r_Controller->muteNodeAddition(true);
    DBG("Querying!");
    connectAndQuery();
    r_Controller->muteNodeAddition(false);
    r_Controller->notifyOfNodeAddition(getRootNode());
}

void OSCQueryAddress::close() {
    signalThreadShouldExit();

    // allow the thread 2 seconds to stop cleanly - should be plenty of time.
    stopThread(2000);
}

Location* OSCQueryAddress::getLocation() {
    return r_Location;
}

void OSCQueryAddress::valueChanged(Value& v) {
    OscAddress::valueChanged(v);

    if (v.refersToSameSourceAs(m_ID)) {
        if (r_Location &&
            (r_Location->getLocationType() == LocationType::OSCQuery) &&
            // TODO WHAT IS THIS!
            (count() == 3)) {
            jassertfalse;
            r_Location->setID(this->getID().toString() + "_Location");
        }
    }
    else if (v.refersToSameSourceAs(m_ConnectionStatus)) {
        bool status = m_ConnectionStatus.getValue();

        if (status && !m_Connected) {
            connectAndQuery();
            m_Connected = true;
        }
        else if (!status && m_Connected) {
            disconnect();
            m_Connected = false;
        }
    }
    else if (v.refersToSameSourceAs(m_OutgoingPortOscQuery)) {
        ScopedLock sl(listenersCriticalSection);
        listeners.call(&Listener::outgoingPortOscQuery_Changed, *this, v.getValue());
        triggerQuery();
    }
    else if (v.refersToSameSourceAs(m_IncomingPort)) {
        ScopedLock sl(listenersCriticalSection);
        listeners.call(&Listener::incomingPort_Changed, *this, v.getValue());
        triggerQuery();
    }
    else if (v.refersToSameSourceAs(m_ChosenZeroconfDeviceName)) {
        if (m_ChosenZeroconfDeviceName.toString().compare("Manual") == 0) {
        }
        else {
            reactToSourceChange();
        }
    }
    else if ((r_OSCQueryLocation.get() != nullptr) && v.refersToSameSourceAs(r_OSCQueryLocation->getOutgoingPort())) {
        int outPort = r_OSCQueryLocation->getOutgoingPort().getValue();
        DBG("OUT PORT SET TO:" + outPort);
    }
    else if ((r_OSCQueryLocation.get() != nullptr) && v.refersToSameSourceAs(r_OSCQueryLocation->getOutgoingIP())) {
        triggerQuery();
    }
}

void OSCQueryAddress::triggerQuery() {
    if (!r_Controller->isLoading() && !r_Controller->isExiting()) {
        startThread();
    }
    else {
        auto lambda = [&] {
            triggerAsyncUpdate();
            return true;
        };

        auto m = new ActionMessageWrapper(r_Controller, std::move(lambda));
        r_Controller->addMessageToGuiOutgoingQueue(m);
    }
}

void OSCQueryAddress::handleAsyncUpdate() {
    startThread();
}

String OSCQueryAddress::getQueryHostString() {
    auto outIP = r_OSCQueryLocation->getOutgoingIP().toString();
    auto outPortOscQuery = getOutgoingPortOscQuery().toString();

    String queryHostString = "ws://" + outIP + ":" + outPortOscQuery;

    std::string host = m_ChosenZeroconfDevice.host;

    if (!host.empty()) {
        bool found = false;
        for (auto device : m_ZeroconfDevices) {
            if (device.name.compare(m_ChosenZeroconfDeviceName.toString().toStdString()) == 0) {
                m_ChosenZeroconfDevice = device;
                outIP = device.ip;
                outPortOscQuery = String(device.port);
                found = true;
                break;
            }
        }

        jassert(found);

        host.pop_back();
        queryHostString = String(host) + ":" + outPortOscQuery;
    }

    return queryHostString;
}

void OSCQueryAddress::connectionSettingsChanged(Location&) {
    triggerQuery();
    DBG("triggerQuery");
}

inline void OSCQueryAddress::reactToSourceChange() {
    std::string outIP;
    int outPort;
    bool found = false;

    {
        ScopedLock zsl(zeroconfCriticalSection);

        for (auto device : m_ZeroconfDevices) {
            if (device.name.compare(m_ChosenZeroconfDeviceName.toString().toStdString()) == 0) {
                m_ChosenZeroconfDevice = device;
                outIP = device.ip;
                outPort = device.port;
                found = true;
                break;
            }
        }
    }

    if (found) {
        if (!r_Controller->getCommandManager()->isUndoingOrRedoing()) {
            auto oldIP = r_OSCQueryLocation->getOutgoingIP().toString();
            auto oldPort = r_OSCQueryLocation->getOutgoingPort().getValue();

            disconnect();
            m_Connected = false;

            auto performLambda = [=] {
                r_OSCQueryLocation->setOutgoingIP(outIP);
                r_OSCQueryLocation->setOutgoingPort(outPort);
                setOutgoingPortOscQuery(outPort);
                triggerQuery();
                return true;
            };

            auto undoLambda = [=] {
                r_OSCQueryLocation->setOutgoingIP(oldIP);
                r_OSCQueryLocation->setOutgoingPort(oldPort);
                setOutgoingPortOscQuery(oldPort);
                // triggerQuery(); // TODO: I should re-instate this if also the nodes created in perform are undone...
                return true;
            };

            auto m = new UndoableOperation(r_Controller, performLambda, undoLambda);
            r_Controller->getCommandManager()->scheduleActionFromCore(m);
        }
    }
    else {
        DBG("This shouldn't happen");
        jassertfalse;
    }
}

void OSCQueryAddress::listenToLocation() {
    if (r_Location) {
        r_Location->addListener(this);
        r_Location->addComponentListener(this);

        r_OSCQueryLocation = dynamic_cast<OscQueryLocation*>(r_Location.get());

        r_OSCQueryLocation->getOutgoingIP().addListener(this);
        r_OSCQueryLocation->getOutgoingPort().addListener(this);
    }
}

void OSCQueryAddress::stopListeningToLocation() {
    if (r_Location) {
        r_Location->removeListener(this);
        r_Location->removeComponentListener(this);

        r_OSCQueryLocation->getOutgoingIP().removeListener(this);
        r_OSCQueryLocation->getOutgoingPort().removeListener(this);
    }
}

void OSCQueryAddress::setLocation(Location* in, bool createContainers) {
    if (in != r_Location || !in) {
        auto previousLocation = r_Location;

        auto iac = r_Controller->getFactory()->getZombies()->getInternalAddressCollection();

        if (r_Location) {
            stopListeningToLocation();
        }
        else { // If location was null, I would have added Address to zombie collection.
            iac->remove(this, false);
        }

        r_Location = in;

        if (!r_Location) {
            auto location = r_Controller->getFactory()->getLocation(LocationType::OSCQuery);
            this->add(location);
            r_Location = location;
            location->setID(this->getID().toString() + "_Location");
            triggerQuery();
        }

        r_OSCQueryLocation = dynamic_cast<OscQueryLocation*>(r_Location.get());

        if (!m_IsBeingDeleted &&
            m_Creator == Creator::User &&
            r_Location->getLocationType() == LocationType::OSCQuery) {
            iac->add(this);
        }

        listenToLocation();

        buildAddressPatternInternal();

        m_Listeners.call(&Address::Listener::locationChanged, *this);

        if (createContainers) {
            r_Controller->getStateBroadcaster()->addressLocationChanged(this, previousLocation);
        }

        if (previousLocation) {
            if (previousLocation->getLocationType() == LocationType::OSCQuery) {
                previousLocation->scheduleDeletion();
            }
        }
    }
}

AddressType OSCQueryAddress::getAddressType() {
    return AddressType::OSCQuery;
}

Value& OSCQueryAddress::getEchoOnLoad() {
    return m_EchoOnLoad;
}

void OSCQueryAddress::setEchoOnLoad(bool muteIn) {
    m_EchoOnLoad = muteIn;
}

Value& OSCQueryAddress::getOutgoingPortOscQuery() {
    return m_OutgoingPortOscQuery;
}

void OSCQueryAddress::setOutgoingPortOscQuery(int in) {
    m_OutgoingPortOscQuery = in;
}

Value& OSCQueryAddress::getIncomingPort() {
    return m_IncomingPort;
}

void OSCQueryAddress::setIncomingPort(int in) {
    m_IncomingPort = in;
}

Value& OSCQueryAddress::getChosenZeroconfDevice() {
    return m_ChosenZeroconfDeviceName;
}

void OSCQueryAddress::setChosenZeroconfDevice(const String& chosen) {
    m_ChosenZeroconfDeviceName = chosen;
}

std::vector<OscZeroconfDevice> OSCQueryAddress::getZeroconfDeviceList() const {
    ScopedLock zsl(zeroconfCriticalSection);

    return m_ZeroconfDevices;
}

Value& OSCQueryAddress::getConnectionStatus() {
    return m_ConnectionStatus;
}

void OSCQueryAddress::addListener(Listener* const listener) {
    ScopedLock sl(listenersCriticalSection);
    listeners.add(listener);
}

void OSCQueryAddress::removeListener(Listener* const listener) {
    ScopedLock sl(listenersCriticalSection);
    listeners.remove(listener);
}
