#pragma once

#include "../Addresses/OscAddress.h"

#include "../../Program/NodeTreeBuilder.h"

namespace ossia {
    class value;
    namespace net {
        class parameter_base;
        class node_base;
        class generic_device;
    }

    namespace oscquery {
        class oscquery_mirror_protocol;
    }
}

struct OscZeroconfDevice {
    std::string name;
    std::string host;
    std::string ip;
    int port{};
};

class OSCQueryAddress : public OscAddress,
                        private Thread,
                        private AsyncUpdater {
public:
    OSCQueryAddress(Value& inOutgoingOscQueryPort,
                    Value& inIncomingPort,
                    String id,
                    Controller* c,
                    bool createContainers = true,
                    Creator creator = Creator::User);

    OSCQueryAddress(const OSCQueryAddress& other);

    ~OSCQueryAddress() override;

    OSCQueryAddress& operator=(const OSCQueryAddress& other);
    virtual OSCQueryAddress* clone() const;

    void fetchZeroconfDevices();

    void getIP(OscZeroconfDevice& newDevice);

    Location* getLocation() override;
    void setLocation(Location* loc, bool createContainers = true) override;

    AddressType	getAddressType();

    Value& getEchoOnLoad();
    void setEchoOnLoad(bool muteIn);

    void valueChanged(Value& value) override;

    void triggerQuery();

    void connectionSettingsChanged(Location& /*loc*/) override;

    class Listener {
    public:
        Listener() {}
        virtual ~Listener() {}
        virtual void outgoingPortOscQuery_Changed(OSCQueryAddress& /*c*/, int /*port*/) {}
        virtual void incomingPort_Changed(OSCQueryAddress& /*c*/, int /*port*/) {}
        virtual void zeroconfList_Changed(OSCQueryAddress& /*c*/) {}
    };

    void addListener(Listener* listener);
    void removeListener(Listener* listener);

    Value& getOutgoingPortOscQuery();
    void setOutgoingPortOscQuery(int in);

    Value& getIncomingPort();
    void setIncomingPort(int in);

    Value& getChosenZeroconfDevice();
    void setChosenZeroconfDevice(const String& chosen);

    std::vector<OscZeroconfDevice> getZeroconfDeviceList() const;

    Value& getConnectionStatus();

protected:
    void reactToSourceChange();

    void listenToLocation() override;
    void stopListeningToLocation() override;

    void on_message_cb(const ossia::net::parameter_base& p);
    void on_unhandled_message_cb(const std::string addressPattern, const ossia::value& val);

    Array<var>* bridgeMessageValues(const ossia::value& val, juce::String& tts);

private:
    void postStatusMessage(const char* message);

    void handleAsyncUpdate() override;

    String getQueryHostString();

    void connectAndQuery();
    void disconnect();

    void parseHostInfo(ossia::oscquery::oscquery_mirror_protocol* protocol);

    void run() override;

    void close();

    void node_created_cb(ossia::net::node_base& n);
    void buildNode(const ossia::net::node_base& n);
    void param_created_cb(const ossia::net::parameter_base& p);

    void node_removing_cb(ossia::net::node_base& n);
    void param_removing_cb(const ossia::net::parameter_base& p);

    void buildNodeLevelTB(const ossia::net::node_base& node);

    void getMetadata(const ossia::net::node_base& node,
                     NewNodeMetadata& metadata);

    void processParameter(ossia::net::parameter_base* parameter, 
                          NewNodeMetadata& metadata);

    void extractParameter(ossia::net::parameter_base* parameter, 
                          NewNodeMetadata& metadata);

    void removeRangesIfMinMaxSame(ossia::value& min, 
                                  ossia::value& max, 
                                  NewNodeMetadata& metadata);

    void* Accept(Visitor& visitor) {
        return visitor.visit_OSCQuery_Address(*this);
    }

    Value m_ConnectionStatus; // bool
    bool m_Connected {false};

    CriticalSection zeroconfCriticalSection;
    Value m_ChosenZeroconfDeviceName;
    OscZeroconfDevice m_ChosenZeroconfDevice;
    std::vector<OscZeroconfDevice> m_ZeroconfDevices;

    std::unique_ptr<ossia::net::generic_device> m_Device;

    NewNodeMetadata m_NewNodeMetadata;

    Value m_EchoOnLoad; // bool

    WeakReference<OscQueryLocation> r_OSCQueryLocation;

    ListenerList<Listener> listeners;
    CriticalSection listenersCriticalSection;

    Value m_OutgoingPortOscQuery; // int
    Value m_IncomingPort; // int

    String m_ConnectedHostString;

    bool m_IsBeingDeleted;

    bool m_HasListen {false};

    JUCE_LEAK_DETECTOR(OSCQueryAddress);
};
